package igblan.life;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;

/**
 *
 * @author Paul Chapman
 */
public abstract class LifePattern
{
	protected static final int[] genTable3x3 = buildGenTable3x3();
	
	private static int[] buildGenTable3x3()
	{
		int[] result = new int[1 << 9];
		
		for (int i = 0; i < result.length; i++)
		{
			int t = Integer.bitCount(i);
			result[i] = t == 3 || t == 4 && (i & 0_020) != 0 ? 1 : 0;
		}
		
		return result;
	}
	
	final static Comparator cellComparator = new Comparator<Cell>()
	{
		@Override
		public int compare(Cell c1, Cell c2)
		{
			return c1.y != c2.y ? c1.y - c2.y : c1.x - c2.x;
		}
	};

	private static void addRLEField(StringBuilder builder, int n, char c)
	{
		if (n == 1) builder.append(c);
		else if (n > 1) builder.append(n).append(c);
	}
	
	public static String toFormattedRLE(String rle, Box box, String rule,
										int cols)
	{
		StringBuilder result = new StringBuilder();
		
		result.append("x = ").append(box.width())
			  .append(", y = ").append(box.height())
			  .append(", rule = ").append(rule).append("\n");

		int n;
		
		for (int i = 0; i < rle.length(); i = n)
		{
			n = Math.min(i + cols, rle.length());
			while (Character.isDigit(rle.charAt(n - 1))) n -= 1;
			result.append(rle.substring(i, n)).append("\n");
		}
		
		return result.toString();
	}
	
	public static String toLifeHistoryRLE(LifePattern pattern, LifePattern smear)
	{
		StringBuilder result = new StringBuilder();
		Cell[] pCells = pattern.toCellArray();
		Cell[] sCells = smear.toCellArray();
		Box box = pattern.getBoundingBox().union(smear.getBoundingBox());
		int x0 = box.left();
		int y0 = box.top();
		char currentType = ' ';
		int n = 0;
		int ip = 0;
		int is = 0;

		while (ip < pCells.length || is < sCells.length)
		{
			char newType;
			Cell cell;
			
			if (is == sCells.length ||
				ip < pCells.length &&
					cellComparator.compare(pCells[ip], sCells[is]) < 0)
			{
				cell = pCells[ip++];
				newType = 'A';
			}
			else
			{
				cell = sCells[is++];
				newType = 'B';
			}
			
			int x = cell.x;
			int y = cell.y;

			if (y != y0)
			{
				addRLEField(result, n, currentType);
				addRLEField(result, y - y0, '$');
				x0 = box.left();
				y0 = y;
				n = 0;
			}

			if (x == x0 && newType == currentType) n += 1;
			else
			{
				addRLEField(result, n, currentType);
				addRLEField(result, x - x0, '.');
				n = 1;
			}
			
			currentType = newType;
			x0 = x + 1;
		}
		
		addRLEField(result, n, currentType);
		return result.append('!').toString();
	}
	
	protected Box boundingBox = null;
	
	protected abstract LifePattern newPattern(int x, int y);
	public abstract LifePattern getEmpty();
	public abstract LifePattern or(LifePattern pattern);
	public abstract LifePattern and(LifePattern pattern);
	public abstract LifePattern andNot(LifePattern pattern);
	abstract public LifePattern smudge();
	public abstract LifePattern next();
	public abstract LifePattern translate(int x, int y);
	public abstract void paint(PatternPainter painter, Object o, Box box);
	
	public Box getBoundingBox() {return boundingBox;}
	public boolean isEmpty() {return getBoundingBox().isEmpty();}
	public LifePattern smooch() {return smut().smudge();}
	public LifePattern translate(Cell p) {return translate(p.x, p.y);}
	
	public LifePattern next(int n)
	{
		LifePattern result = this;
		for (int i = 0; i < n; i++) result = result.next();
		return result;
	}
	
	public LifePattern smudge(int n)
	{
		LifePattern result = this;
		for (int i = 0; i < n; i++) result = result.smudge();
		return result;
	}

	public LifePattern smut()
	{
		return or(translate(0, -1)).or(translate(-1, 0)).or(translate(1, 0)).
			   or(translate(0, 1));
	}

	public LifePattern newPatternFromRLE(String s)
	{
		LifePattern result = getEmpty();
		
		int x = 0;
		int y = 0;
		int n = 0;
		
		for (int i = 0; i < s.length(); i++)
		{
			char c = s.charAt(i);
			
			if (Character.isDigit(c)) n = 10 * n + Character.digit(c, 10);
			else if (c == 'b' || c == 'o' || c == '$')
			{
				n = n != 0 ? n : 1;
				if (c == 'b')
				{
					x += n;
					n = 0;
				}
				else if (c == '$')
				{
					y += n;
					x = 0;
					n = 0;
				}
				else if (c == 'o')
				{
					for (int x0 = x; x0 < x + n; x0++)
						result = result.or(newPattern(x0, y));

					x += n;
					n = 0;
				}
			}
		}

		return result;
	}
	
	public LifePattern[] smear(int n)
	{
		LifePattern result = this;
		LifePattern evenSmear = this;
		LifePattern oddSmear = getEmpty();
		
		for (int i = 0; i < n; i++)
		{
			result = result.next();
			if ((i & 1) == 0) oddSmear = oddSmear.or(result);
			else evenSmear = evenSmear.or(result);
		}
		
		return new LifePattern[] {result, evenSmear.or(oddSmear)};
	}
	
	public Cell[] toCellArray()
	{
		final ArrayList<Cell> list = new ArrayList();
		
		PatternPainter painter = new PatternPainter()
		{
			@Override
			public void paintCell(Object o, int x, int y)
			{
				list.add(new Cell(x, y));
			}
		};

		paint(painter, null, getBoundingBox());

		Cell[] result = list.toArray(new Cell[0]);
		
		Arrays.sort(result, cellComparator);
		
		return result;
	}
	
	public String toRawRLE()
	{
		StringBuilder result = new StringBuilder();
		Cell[] cells = toCellArray();
		int x0 = getBoundingBox().left();
		int y0 = getBoundingBox().top();
		int n = 0;
		
		for (int i = 0; i < cells.length; i++)
		{
			Cell cell = cells[i];
			int x = cell.x;
			int y = cell.y;
			
			if (y != y0)
			{
				addRLEField(result, n, 'o');
				addRLEField(result, y - y0, '$');
				x0 = getBoundingBox().left();
				y0 = y;
				n = 0;
			}
			
			if (x == x0) n += 1;
			else
			{
				addRLEField(result, n, 'o');
				addRLEField(result, x - x0, 'b');
				n = 1;
			}
			
			x0 = x + 1;
		}
		
		addRLEField(result, n, 'o');
		return result.append('!').toString();
	}
	
	public String toFormattedRLE(int cols)
	{
		return toFormattedRLE(toRawRLE(), getBoundingBox(), "B3/S23", cols);
	}

	public LifePattern homed()
	{
		return translate(getBoundingBox().leftTop.negated());
	}
}
