package igblan.life;

import java.util.Arrays;

public class RunPattern extends LifePattern
{
	private static final int SIDE = 227016;
	private static final int WIDTHINBRICKS = SIDE / 12; // 18918
	private static final int HEIGHTINBRICKS = SIDE / 2; // 113508
	private static final int TOTALCOUNT =
		WIDTHINBRICKS * HEIGHTINBRICKS; // 0x7FFDDFD8
	private static final int TABLEBITS = 16;
	private static final int LOOKUPBITS = TABLEBITS / 2; // 8
	private static final int TABLESIZE = (1 << TABLEBITS); // 65536
	private static final int LOMASK = (1 << LOOKUPBITS) - 1; // 0x00FF
	private static final int HIMASK = LOMASK << LOOKUPBITS; // 0xFF00

	private static class WriteStream
	{
		protected int[] buffer;
		protected int ptr = 0;
		protected int count = 0;

		WriteStream(int length)
		{
			buffer = new int[Math.max(128, 2 * length)];
			buffer[ptr++] = 0;
		}

		private void put(int i, int n)
		{
			if (i >= buffer.length)
				buffer = Arrays.copyOf(buffer, 2 * buffer.length);
			
			buffer[i] = n;
		}
		
		void append(int word)
		{
			if (word == 0) count += 1;
			else
			{
				if (count != 0)
				{
					buffer[ptr - 1] |= 1 << 30;
					put(ptr++, count);
				}

				put(ptr++, word);
				count = 0;
			}
		}

		void skip(int n) {count += n;}

		int[] closeBuffer()
		{
			buffer[ptr - 1] |= 3 << 30;
			put(ptr++, count);
			put(ptr++, 0);
			return Arrays.copyOf(buffer, ptr);
		}
	}

	private static final Box[] boxTable = buildBoxes();

	private static Box[] buildBoxes()
	{
		Box[] result = new Box[4096];
		
		for (int word = 1; word < 4096; word++)
		{
			int left = 100;
			int top = 100;
			int right = -100;
			int bottom = -100;
			
			for (int y = 0; y < 2; y++)
				for (int x = 0; x < 6; x++)
					if ((word & 1 << x + x + y) != 0)
					{
						left = Math.min(left, x);
						top = Math.min(top, y);
						right = Math.max(right, x + 1);
						bottom = Math.max(bottom, y + 1);
					}
			
			result[word] = new Box(left, top, right, bottom);
		}
		
		return result;
	}
	
	private static final int[] genTable = buildGenTable();
	private static final int[] genTopTable = buildGenTopTable();
	private static final int[] genBottomTable = buildGenBottomTable();

	private static final int[] smudgeTable = buildSmudgeTable();
	private static final int[] smudgeTopTable = buildSmudgeTopTable();
	private static final int[] smudgeBottomTable = buildSmudgeBottomTable();
	
	private static class Tables
	{
		public final int[] main;
		public final int[] top;
		public final int[] bottom;

		public Tables(int[] main, int[] top, int[] bottom)
		{
			this.main = main;
			this.top = top;
			this.bottom = bottom;
		}
	}
	
	private static final Tables genTables =
		new Tables(genTable, genTopTable, genBottomTable);
	private static final Tables smudgeTables =
		new Tables(smudgeTable, smudgeTopTable, smudgeBottomTable);
	
	// ..............
	// .FFEEDDCCBBAA.
	// .FFEEDDCCBBAA.
	// ..............

	// MKIGECA86420
	// NLJHFDB97531
	
	// xr yr x y xBit yBit 
	
	private static int[] buildGenTable()
	{
		int[] result = new int[TABLESIZE]; // 65536
		final int XLIMIT = LOOKUPBITS / 2 - 2; // 2

		for (int i = 0; i < TABLESIZE; i++)
		{
			int nr = 0;
			int mr = 1;

			for (int xr = 1; xr <= XLIMIT; xr++)
				for (int yr = 1; yr <= 2; yr++)
				{
					int t = 0;
					int c = 0;

					for (int y = yr - 1; y <= yr + 1; y++)
						for (int x = xr - 1; x <= xr + 1; x++)
						{
							int xBit = x << 1; // 0 2 4 6
							int yBit = y < 2 ? y + LOOKUPBITS : y - 2;
								// 8 9 0 1
							int b = i >> xBit + yBit & 1;
								// E C A 8
								// F D B 9
								// 6 4 2 0
								// 7 5 3 1
							if (x == xr && y == yr) c = b;
							t += b;
						}

					if (t == 3 || (t == 4 && c != 0)) nr |= mr;

					mr <<= 1;
				}

			result[i] = nr;
		}
		
		return result;
	}

	private static int[] buildGenTopTable()
	{
		int[] result = new int[TABLESIZE];
		
		for (int i = 0; i < TABLESIZE; i++)
		{
			int nr = 0;
			int mr = 1;

			for (int xr = 1; xr <= 6; xr++)
				for (int yr = 1; yr <= 2; yr++)
				{
					int t = 0;
					int c = 0;

					for (int y = yr - 1; y <= yr + 1; y++)
						for (int x = xr - 1; x <= xr + 1; x++)
						{
							int b = y < 2 ? i >> x + x + y & 1 : 0;
							if (x == xr && y == yr) c = b;
							t += b;
						}

					if (t == 3 || (t == 4 && c != 0)) nr |= mr;

					mr <<= 1;
				}

			result[i] = nr;
		}
		
		return result;
	}

	private static int[] buildGenBottomTable()
	{
		int[] result = new int[TABLESIZE];
		
		for (int i = 0; i < TABLESIZE; i++)
		{
			int nr = 0;
			int mr = 1;

			for (int xr = 1; xr <= 6; xr++)
				for (int yr = 1; yr <= 2; yr++)
				{
					int t = 0;
					int c = 0;

					for (int y = yr - 1; y <= yr + 1; y++)
						for (int x = xr - 1; x <= xr + 1; x++)
						{
							int b = y < 2 ? 0 : (i >> (x + x + y - 2)) & 1;
							if (x == xr && y == yr) c = b;
							t += b;
						}

					if (t == 3 || (t == 4 && c != 0)) nr |= mr;

					mr <<= 1;
				}

			result[i] = nr;
		}
		
		return result;
	}

	private static int[] buildSmudgeTable()
	{
		int[] result = new int[TABLESIZE];
		final int XLIMIT = LOOKUPBITS / 2 - 2; // 2

		for (int i = 0; i < TABLESIZE; i++)
		{
			int nr = 0;
			int mr = 1;

			for (int xr = 1; xr <= XLIMIT; xr++)
				for (int yr = 1; yr <= 2; yr++)
				{
					int t = 0;

					for (int y = yr - 1; y <= yr + 1; y++)
						for (int x = xr - 1; x <= xr + 1; x++)
						{
							int xBit = x << 1;
							int yBit = y < 2 ? y + LOOKUPBITS : y - 2;
							int b = i >> xBit + yBit & 1;
							t += b;
						}

					if (t != 0) nr |= mr;

					mr <<= 1;
				}

			result[i] = nr;
		}
		
		return result;
	}

	private static int[] buildSmudgeTopTable()
	{
		int[] result = new int[TABLESIZE];
		
		for (int i = 0; i < TABLESIZE; i++)
		{
			int nr = 0;
			int mr = 1;

			for (int xr = 1; xr <= 6; xr++)
				for (int yr = 1; yr <= 2; yr++)
				{
					int t = 0;

					for (int y = yr - 1; y <= yr + 1; y++)
						for (int x = xr - 1; x <= xr + 1; x++)
						{
							int b = y < 2 ? i >> x + x + y & 1 : 0;
							t += b;
						}

					if (t != 0) nr |= mr;

					mr <<= 1;
				}

			result[i] = nr;
		}
		
		return result;
	}

	private static int[] buildSmudgeBottomTable()
	{
		int[] result = new int[TABLESIZE];
		
		for (int i = 0; i < TABLESIZE; i++)
		{
			int nr = 0;
			int mr = 1;

			for (int xr = 1; xr <= 6; xr++)
				for (int yr = 1; yr <= 2; yr++)
				{
					int t = 0;

					for (int y = yr - 1; y <= yr + 1; y++)
						for (int x = xr - 1; x <= xr + 1; x++)
						{
							int b = y < 2 ? 0 : i >> x + x + y - 2 & 1;
							t += b;
						}

					if (t != 0) nr |= mr;

					mr <<= 1;
				}

			result[i] = nr;
		}
		
		return result;
	}

	public static final RunPattern empty = new RunPattern();
	
	private final int[] buffer;
	private final boolean jogged;

	abstract private class AbstractReadStream
	{
		abstract int count();
		abstract int next();
		abstract void quickSkip(int n);
		abstract boolean atEnd();
	}
	
	private class ReadStream extends AbstractReadStream
	{
		protected int ptr;
		protected int countv;

		ReadStream()
		{
			ptr = 1;
			countv = buffer[ptr++];
		}

		@Override
		int count() {return countv;}

		@Override
		int next()
		{
			if (countv != 0)
			{
				--countv;
				return 0;
			}
			else
			{
				int word = buffer[ptr++];
				int flags = word >>> 30;

				if (flags != 0)
				{
					countv = buffer[ptr++];
					word &= 0x0FFFFFF0;
				}

				return word;
			}
		}

		@Override
		void quickSkip(int n) {countv -= n;}

		@Override
		boolean atEnd() {return countv == 0 && buffer[ptr] == 0;}
	}
	
	private class ReadStreamCarryForward extends ReadStream
	{
		private int carry = 0;
		
		ReadStreamCarryForward()
		{
			super();
		}
		
		@Override
		int count() {return carry != 0 ? 0 : countv;}

		@Override
		int next()
		{
			if (countv != 0)
			{
				int temp = carry;
				carry = 0;
				--countv;
				return temp;
			}
			else
			{
				int word = buffer[ptr++];
				
				if (word >>> 30 != 0)
				{
					countv = buffer[ptr++];
					word &= 0x0FFFFFF0;
				}

				int temp = carry;
				carry = word >>> 24;
				return word | temp;
			}
		}

		@Override
		void quickSkip(int n)
		{
			super.quickSkip(n);
			if (n != 0) carry = 0;
		}
	}
	
	private class ReadStreamCarryBackward extends ReadStream
	{
		ReadStreamCarryBackward()
		{
			super();
		}

		@Override
		int count()
		{
			return countv != 0 && (buffer[ptr] & 0xF0) != 0 ?
				countv - 1 : countv;
		}

		@Override
		int next()
		{
			if (countv != 0)
			{
				if (--countv == 0) return (buffer[ptr] & 0xF0) << 20;
				else return 0;
			}
			else
			{
				int word = buffer[ptr++];

				if (word >>> 30 != 0)
				{
					countv = buffer[ptr++];
					word &= 0x0FFFFFF0;
				}
				else word |= (buffer[ptr] & 0xF0) << 24;

				return word >>> 4;
			}
		}
	}

	abstract private class AbstractReadStreamJog extends AbstractReadStream
	{
		protected ReadStream top;
		protected ReadStream bottom;
		
		@Override
		int count()
		{
			if (top.count() == 0 || bottom.count() == 0) return 0;
			else return Math.min(top.count(), bottom.count());
		}
		
		@Override
		int next()
		{
			if (top.count() != 0)
			{
				if (bottom.count() != 0)
				{
					top.quickSkip(1);
					bottom.quickSkip(1);
					return 0;
				}
				else
				{
					top.quickSkip(1);
					return bottom.next() << 3 & 0x0AAAAAA0;
				}
			}
			else
			{
				if (bottom.count() != 0)
				{
					bottom.quickSkip(1);
					return top.next() << 1 & 0x05555550;
				}
				else
				{
					return top.next() << 1 & 0x05555550 |
						   bottom.next() << 3 & 0x0AAAAAA0;
				}
			}
		}
		
		@Override
		void quickSkip(int n)
		{
			top.quickSkip(n);
			bottom.quickSkip(n);
		}
		
		@Override
		boolean atEnd()
		{
			return bottom.atEnd();
		}
	}
	
	private class ReadStreamJog extends AbstractReadStreamJog
	{
		ReadStreamJog()
		{
			top = new ReadStreamCarryForward();
			bottom = new ReadStreamCarryForward();
			bottom.quickSkip(SIDE / 12);
		}
	}
	
	private class ReadStreamUnjog extends AbstractReadStreamJog
	{
		ReadStreamUnjog()
		{
			top = new ReadStreamCarryBackward();
			bottom = new ReadStreamCarryBackward();
			bottom.quickSkip(SIDE / 12);
		}
	}
	
	private class ReadStreamXShifted extends ReadStream
	{
		private int offset;
		private int prev;

		ReadStreamXShifted(int xoffset)
		{
			super();
			offset = xoffset * 2;
			prev = 0;
		}

		@Override
		int count()
		{
			return prev != 0 ? 0 : countv;
		}

		@Override
		int next()
		{
			if (countv != 0)
			{
				int temp = prev;
				prev = 0;
				--countv;
				return temp;
			}
			else
			{
				int word = buffer[ptr++];

				if (word >>> 30 != 0)
				{
					countv = buffer[ptr++];
					word &= 0x0FFFFFF0;
				}

				int temp = prev;
				prev = (word >>> (24 - offset)) & 0x0FFFFFF0;
				return ((word << offset) & 0x0FFFFFF0) | temp;
			}
		}

		@Override
		void quickSkip(int n)
		{
			super.quickSkip(n);
			if (n != 0) prev = 0;
		}
	}

	private static RunPattern fromRLE(String s)
	{
		return (RunPattern) empty.newPatternFromRLE(s);
	}
	
	private RunPattern()
	{
		buffer = new int[3];
		
		buffer[0] = 3 << 30;
		buffer[1] = TOTALCOUNT;
		buffer[2] = 0;
		
		jogged = false;
	}
	
	private RunPattern(LifePattern p)
		{this(((RunPattern) p).buffer, ((RunPattern) p).jogged);}
	
	public RunPattern(String s){this(empty.newPatternFromRLE(s));}
	
	public RunPattern(int x, int y)
	{
		buffer = new int[5];
		
		int offset = SIDE >> 1;
		int x0 = x + offset;
		int y0 = y + offset;
		int r = x0 % 12;
		int c = (y0 >> 1) * (SIDE / 12) + x0 / 12;
		int s = (r << 1) | (y0 & 1);
		int m = 0x10 << s;

		buffer[0] = 2 << 30;
		buffer[1] = c;
		buffer[2] = m | (3 << 30);
		buffer[3] = TOTALCOUNT - c - 1;
		buffer[4] = 0;
		
		jogged = false;
	}
	
	RunPattern(int[] buffer, boolean jogged)
	{
		this.buffer = buffer;
		this.jogged = jogged;
	}
	
	RunPattern(int[] buffer, boolean jogged, Box boundingBox)
	{
		this.buffer = buffer;
		this.jogged = jogged;
		this.boundingBox = boundingBox;
	}
	
	@Override
	protected LifePattern newPattern(int x, int y)
	{
		return new RunPattern(x, y);
	}
	
	void jogCommon(WriteStream result, ReadStream top, ReadStream bottom)
	{
		while (!bottom.atEnd())
		{
			if (top.count() != 0)
			{
				if (bottom.count() != 0)
				{
					int n = Math.min(top.count(), bottom.count());

					top.quickSkip(n);
					bottom.quickSkip(n);
					result.skip(n);
				}
				else
				{
					top.quickSkip(1);
					result.append(bottom.next() << 3 & 0x0AAAAAA0);
				}
			}
			else
			{
				if (bottom.count() != 0)
				{
					bottom.quickSkip(1);
					result.append(top.next() << 1 & 0x05555550);
				}
				else
				{
					result.append(top.next() << 1 & 0x05555550 |
								  bottom.next() << 3 & 0x0AAAAAA0);
				}
			}
		}
	}
	
	private RunPattern jog()
	{
		WriteStream w = new WriteStream(buffer.length);

		if (!jogged)
		{
			ReadStream top = new ReadStreamCarryForward();
			ReadStream bottom = new ReadStreamCarryForward();

			bottom.quickSkip(SIDE / 12);
			w.skip(SIDE / 12);
			jogCommon(w, top, bottom);
		}
		else
		{
			ReadStream top = new ReadStreamCarryBackward();
			ReadStream bottom = new ReadStreamCarryBackward();

			bottom.quickSkip(SIDE / 12);
			jogCommon(w, top, bottom);
			w.skip(SIDE / 12);
		}

		return new RunPattern(w.closeBuffer(), !jogged, boundingBox);
	}
	
	private RunPattern jogIf(boolean b) {return b ? jog() : this;}

	@Override
	public int hashCode() {return Arrays.hashCode(jogIf(jogged).buffer);}

	@Override
	public boolean equals(Object object)
	{
		if (object == null) return false;
		if (getClass() != object.getClass()) return false;
		
		RunPattern pattern = (RunPattern) object;
		
		return Arrays.equals(jogIf(jogged != pattern.jogged).buffer,
							 pattern.buffer);
	}
	
	private void next(WriteStream w, ReadStream top, ReadStream bottom,
					  Tables tables)
	{
		while (!bottom.atEnd())
		{
			if (top.count() != 0)
			{
				if (bottom.count() != 0)
				{
					int n = Math.min(top.count(), bottom.count());

					top.quickSkip(n);
					bottom.quickSkip(n);
					w.skip(n);
				}
				else
				{
					top.quickSkip(1);

					int b = bottom.next();
					int n = 0;

					n |= tables.bottom[b & 0xFFFF] << 4;
					b = Integer.rotateRight(b, 12);
					n |= tables.bottom[b & 0xFFFF] << 16;

					w.append(n);
				}
			}
			else
			{
				if (bottom.count() != 0)
				{
					bottom.quickSkip(1);

					int a = top.next();
					int n = 0;

					n |= tables.top[a & 0xFFFF] << 4;
					a = Integer.rotateRight(a, 12);
					n |= tables.top[a & 0xFFFF] << 16;

					w.append(n);
				}
				else
				{
					int a = Integer.rotateLeft(top.next(), LOOKUPBITS);
					int b = bottom.next();
					int n = 0;
					
					// 7-0
					n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 4;
					a = Integer.rotateRight(a, 4);
					b = Integer.rotateRight(b, 4);
					// 11-4
					n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 8;
					a = Integer.rotateRight(a, 4);
					b = Integer.rotateRight(b, 4);
					// 15-8
					n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 12;
					a = Integer.rotateRight(a, 4);
					b = Integer.rotateRight(b, 4);
					// 19-12
					n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 16;
					a = Integer.rotateRight(a, 4);
					b = Integer.rotateRight(b, 4);
					// 23-16
					n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 20;
					a = Integer.rotateRight(a, 4);
					b = Integer.rotateRight(b, 4);
					// 27-20
					n |= tables.main[(a & HIMASK) | (b & LOMASK)] << 24;

					w.append(n);
				}
			}
		}
	}
	
	private RunPattern next(Tables tables)
	{
		WriteStream w = new WriteStream(buffer.length);
		
		if (!jogged)
		{
			ReadStream top = new ReadStreamCarryForward();
			ReadStream bottom = new ReadStreamCarryForward();

			bottom.quickSkip(SIDE / 12);
			w.skip(SIDE / 12);
			next(w, top, bottom, tables);
		}
		else
		{
			ReadStream top = new ReadStreamCarryBackward();
			ReadStream bottom = new ReadStreamCarryBackward();

			bottom.quickSkip(SIDE / 12);
			next(w, top, bottom, tables);
			w.skip(SIDE / 12);
		}
		
		return new RunPattern(w.closeBuffer(), !jogged);
	}

	@Override
	public LifePattern next() {return next(genTables);}
	
	@Override
	public LifePattern smudge() {return next(smudgeTables);}
	
	public LifePattern orNEW(LifePattern pattern)
	{
		if (isEmpty()) return pattern;
		if (pattern.isEmpty()) return this;
		
		RunPattern runPattern = (RunPattern) pattern;
		ReadStream r1 = new ReadStream();
		AbstractReadStream r2 =
			jogged & !runPattern.jogged ? runPattern.new ReadStreamJog() :
			runPattern.jogged &! jogged ? runPattern.new ReadStreamUnjog() :
			new ReadStream();
		WriteStream w = new WriteStream(Math.max(buffer.length,
												 runPattern.buffer.length));

		while (!r1.atEnd())
		{
			if (r1.count() != 0)
			{
				if (r2.count() != 0)
				{
					int n = Math.min(r1.count(), r2.count());
					r1.quickSkip(n);
					r2.quickSkip(n);
					w.skip(n);
				}
				else
				{
					r1.quickSkip(1);
					w.append(r2.next());
				}
			}
			else
			{
				if (r2.count() != 0)
				{
					r2.quickSkip(1);
					w.append(r1.next());
				}
				else w.append(r1.next() | r2.next());
			}
		}

		if (jogged ^ runPattern.jogged) w.skip(SIDE / 12);
		
		return new RunPattern(w.closeBuffer(), jogged,
							  getBoundingBox().union(pattern.getBoundingBox()));
	}
	
	@Override
	public LifePattern or(LifePattern pattern)
	{
		if (isEmpty()) return pattern;
		if (pattern.isEmpty()) return this;
		
		RunPattern runPattern = (RunPattern) pattern;
		ReadStream r1 = new ReadStream();
		ReadStream r2 =
				runPattern.jogIf(jogged ^ runPattern.jogged).new ReadStream();
		WriteStream w = new WriteStream(Math.max(buffer.length,
												 runPattern.buffer.length));

		while (!r1.atEnd())
		{
			if (r1.count() != 0)
			{
				if (r2.count() != 0)
				{
					int n = Math.min(r1.count(), r2.count());
					r1.quickSkip(n);
					r2.quickSkip(n);
					w.skip(n);
				}
				else
				{
					r1.quickSkip(1);
					w.append(r2.next());
				}
			}
			else
			{
				if (r2.count() != 0)
				{
					r2.quickSkip(1);
					w.append(r1.next());
				}
				else w.append(r1.next() | r2.next());
			}
		}

		return new RunPattern(w.closeBuffer(), jogged,
							  getBoundingBox().union(pattern.getBoundingBox()));
//		RunPattern result = new RunPattern(w.closeBuffer(), jogged,
//							  getBoundingBox().union(pattern.getBoundingBox()));
//		if (!result.equals(orNEW(pattern))) System.exit(1);
//		return result;
	}
	
	@Override
	public LifePattern and(LifePattern pattern)
	{
		if (isEmpty() || pattern.isEmpty()) return empty;
		
		RunPattern runPattern = (RunPattern) pattern;
		ReadStream r1 = new ReadStream();
		ReadStream r2 =
				runPattern.jogIf(jogged ^ runPattern.jogged).new ReadStream();
		WriteStream w = new WriteStream(Math.max(buffer.length,
												 runPattern.buffer.length));

		while (!r1.atEnd())
		{
			if (r1.count() != 0)
			{
				if (r2.count() != 0)
				{
					int n = Math.min(r1.count(), r2.count());
					r1.quickSkip(n);
					r2.quickSkip(n);
					w.skip(n);
				}
				else
				{
					r1.quickSkip(1);
					r2.next();
					w.append(0);
				}
			}
			else
			{
				if (r2.count() != 0)
				{
					r2.quickSkip(1);
					r1.next();
					w.append(0);
				}
				else w.append(r1.next() & r2.next());
			}
		}

		return new RunPattern(w.closeBuffer(), jogged);
	}

	@Override
	public LifePattern andNot(LifePattern pattern)
	{
		if (isEmpty()) return empty;
		
		RunPattern runPattern = (RunPattern) pattern;
		ReadStream r1 = new ReadStream();
		ReadStream r2 =
				runPattern.jogIf(jogged ^ runPattern.jogged).new ReadStream();
		WriteStream w = new WriteStream(Math.max(buffer.length,
												 runPattern.buffer.length));

		while (!r1.atEnd())
		{
			if (r1.count() != 0)
			{
				if (r2.count() != 0)
				{
					int n = Math.min(r1.count(), r2.count());
					r1.quickSkip(n);
					r2.quickSkip(n);
					w.skip(n);
				}
				else
				{
					r1.quickSkip(1);
					r2.next();
					w.append(0);
				}
			}
			else
			{
				if (r2.count() != 0)
				{
					r2.quickSkip(1);
					w.append(r1.next());
				}
				else w.append(r1.next() &~ r2.next());
			}
		}

		return new RunPattern(w.closeBuffer(), jogged);
	}

	@Override
	public LifePattern translate(int x, int y)
	{
		if (isEmpty()) return this;
		
		int adjust = (y & 1) == 0 ? 0 : jogged ? -1 : 1;
		int xpos = x + adjust + SIDE / 2;
		int xoffset = xpos / 12 - WIDTHINBRICKS / 2;
		int yoffset = (y + adjust) * (WIDTHINBRICKS / 2);
		int offset = xoffset + yoffset;

		ReadStreamXShifted r = new ReadStreamXShifted(xpos % 12);
		WriteStream w = new WriteStream(buffer.length);

		w.skip(offset);

		while (true)
		{
			int count = r.count();
			r.quickSkip(count);
			w.skip(count);

			if (r.atEnd()) break;

			while (true)
			{
				int word = r.next();
				w.append(word);
				if (word == 0) break;
			}
		}

		w.skip(- offset);

		return new RunPattern(w.closeBuffer(), jogged ^ (y & 1) != 0,
							  getBoundingBox().translate(x, y));
	}

	@Override
	public Box getBoundingBox()
	{
		if (boundingBox != null) return boundingBox;
		
		int offset = SIDE / 2 + (jogged ? 1 : 0);
		ReadStream r = new ReadStream();
		int left = SIDE / 2;
		int top = SIDE / 2;
		int right = - SIDE / 2;
		int bottom = - SIDE / 2;
		int position = 0;
		boolean isEmpty = true;

		while (true)
		{
			int count = r.count();
			r.quickSkip(count);
			position += count;

			if (r.atEnd()) break;

			for (int word = r.next(); word != 0; word = r.next())
			{
				int x = (int) ((position % WIDTHINBRICKS) * 12) - offset;
				int y = (int) ((position / WIDTHINBRICKS) * 2) - offset;
				position++;
				
				int mask;
				
				mask = word >> 4 & 0xFFF;
				
				if (mask != 0)
				{
					Box box = boxTable[mask];
					left = Math.min(left, x + box.left());
					top = Math.min(top, y + box.top());
					right = Math.max(right, x + box.right());
					bottom = Math.max(bottom, y + box.bottom());
				}

				mask = word >> 16 & 0xFFF;
				
				if (mask != 0)
				{
					Box box = boxTable[mask];
					int x6 = x + 6;
					left = Math.min(left, x6 + box.left());
					top = Math.min(top, y + box.top());
					right = Math.max(right, x6 + box.right());
					bottom = Math.max(bottom, y + box.bottom());
				}

				isEmpty = false;
			}
			
			position++;
		}

		boundingBox = isEmpty ? Box.empty : new Box(left, top, right, bottom);
		
		return boundingBox;
	}
	
	@Override
	public LifePattern getEmpty() {return empty;}

	@Override
	public String toString()
	{
		return "placeholder";
	}
	
	@Override
	public void paint(PatternPainter painter, Object o, Box box)
	{
		int offset = SIDE / 2 + (jogged ? 1 : 0);
		ReadStream r = new ReadStream();
		int position = 0;

		while (true)
		{
			{
				int count = r.count();
				r.quickSkip(count);
				position += count;
			}

			if (r.atEnd()) break;

			for (int word = r.next(); word != 0; word = r.next())
			{
				int x0 = 12 * (position % WIDTHINBRICKS) - offset;
				int y0 = 2 * (position / WIDTHINBRICKS) - offset;
				position++;
				
				if (!box.intersects(new Box(x0, y0, x0 + 12, y0 + 2)))
					continue;
				
				for (int y = 0; y < 2; y++)
					for (int x = 0; x < 12; x++)
					{
						int x1 = x0 + x;
						int y1 = y0 + y;

						if (box.contains(x1, y1))
							if ((word & 0x10 << x + x + y) != 0)
								painter.paintCell(o, x1, y1);
					}
			}
			
			position++;
		}
	}
}
